/***********************************************************************************
 *   Copyright (C) 2009 Valerio Orlandini <valerio.orlandini@gmx.com>
 *   The base for this work comes from the amplifier example, written by
 *   Steve Harris <http://plugin.org.uk/>, the noise generators from the work of
 * 	 Nick Lamb <njl195@zepler.org.uk> and James W. Morris <james@jwm-art.net>
 * 
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *   Questo programma è software libero, soggetto ai termini della licenza
 *   GNU General Public License. Può essere modificato e ridistribuito
 *   secondo i termini specificati.
 *   Il presente programma è distribuito senza garanzia alcuna.
 *********************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <stdio.h>

#include "include/lv2.h"
#include "omnifx.c"

/* Defining each port in order to make the code clearer */
#define OMF_URI "http://interferon.drivehq.com/plugins/omnifono";
#define INPUT_1 0
#define INPUT_2 1
#define OUTPUT_1 2
#define OUTPUT_2 3
#define GAIN_L 4
#define GAIN_R 5
#define PAN 6
#define CROSS_RL 7
#define CROSS_LR 8
#define NOIS1_VOL 9
#define NOIS2_VOL 10
#define ANGLE 11

static LV2_Descriptor *omfDescriptor = NULL;

/* Structure of the effect */
typedef struct {
	float *gain_l;
	float *gain_r;
	float *panner;
	float *input_l;
	float *input_r;
	float *output_l;
	float *output_r;
	float *cross_rl;
	float *cross_lr;
	float *noise1_vol;
	float *noise2_vol;
	float *angle;
	double sample_rate;
} omf;

static void cleanupOmf(LV2_Handle instance)
{
	free(instance);
}

/* Data to be sent for each port */
static void connectPortOmf(LV2_Handle instance, uint32_t port,
			   void *data)
{
	omf *plugin = (omf *)instance;

	switch (port) {
	case INPUT_1:
		plugin->input_l = data;
		break;
	case INPUT_2:
		plugin->input_r = data;
		break;
	case OUTPUT_1:
		plugin->output_l = data;
		break;
	case OUTPUT_2:
		plugin->output_r = data;
		break;
	case GAIN_L:
		plugin->gain_l = data;
		break;
	case GAIN_R:
		plugin->gain_r = data;
		break;
	case PAN:
		plugin->panner = data;
		break;
	case CROSS_RL:
		plugin->cross_rl = data;
		break;
	case CROSS_LR:
		plugin->cross_lr = data;
		break;
	case NOIS1_VOL:
		plugin->noise1_vol = data;
		break;
	case NOIS2_VOL:
		plugin->noise2_vol = data;
		break;
	case ANGLE:
		plugin->angle = data;
		break;
	}
}

static LV2_Handle instantiate   (
                                                    const LV2_Descriptor * descriptor,
                                                    double SampleRate,
                                                    const char * bundle_path,
                                                    const LV2_Feature * const * host_features
                                                    )
{
	omf *plugin_data = (omf *)malloc(sizeof(omf));
	return (LV2_Handle)plugin_data;
}

#define DB_CO(g) ((g) > -90.0f ? powf(10.0f, (g) * 0.05f) : 0.0f)

static void runOmf(LV2_Handle instance, uint32_t sample_count)
{
	omf *plugin_data = (omf *)instance;
    
	/* Some definitions to shorten the following code */
	const float gain_l = *(plugin_data->gain_l);
	const float gain_r = *(plugin_data->gain_r);
	const float * const inl = plugin_data->input_l;
	const float * const inr = plugin_data->input_r;
	float * const outl = plugin_data->output_l;
	float * const outr = plugin_data->output_r;
	float pan = *(plugin_data->panner) / 100;
	float crl = *(plugin_data->cross_rl) / 100;
	float clr = *(plugin_data->cross_lr) / 100;
	float n1v = *(plugin_data->noise1_vol) / 100;
	float n2v = *(plugin_data->noise2_vol) / 100;
        float angle = *(plugin_data->angle) / (360 / 2 * M_PI);
	float coef_l = DB_CO(gain_l);
	float coef_r = DB_CO(gain_r);
	
	/* Et voilà... effect processing is demanded to function sndgen(), in the omnifx.c
	 * file, then equalized through the function do_3band() in the eq.c file */
	register uint32_t pos;
        int infol = 1;
        int infor = 0;
	for (pos = 0; pos < sample_count; pos++) {
		outl[pos] = sndgen(inl[pos], inr[pos], crl, (1-pan), coef_l, n1v, n2v, angle, infol);
		outr[pos] = sndgen(inr[pos], inl[pos], clr, pan, coef_r, n1v, n2v, angle, infor);
	}
}

/* LV2 plugin mandatory part */
static void init()
{
	omfDescriptor =
	 (LV2_Descriptor *)malloc(sizeof(LV2_Descriptor));

	omfDescriptor->URI = OMF_URI;
	omfDescriptor->activate = NULL;
	omfDescriptor->cleanup = cleanupOmf;
	omfDescriptor->connect_port = connectPortOmf;
	omfDescriptor->deactivate = NULL;
	omfDescriptor->instantiate = instantiate;
	omfDescriptor->run = runOmf;
	omfDescriptor->extension_data = NULL;
}

LV2_SYMBOL_EXPORT
const LV2_Descriptor *lv2_descriptor(uint32_t index)
{
	if (!omfDescriptor) init();

	switch (index) {
	case 0:
		return omfDescriptor;
	default:
		return NULL;
	}
}
